import requests

from json.decoder import JSONDecodeError

from .exceptions import ClientLoginRequired, ClientError
from .utils import gen_csrftoken


class Account:

    def reset_password(self, username):
        response = requests.post(
            "https://www.instagram.com/accounts/account_recovery_send_ajax/",
            data={
                "email_or_username": username,
                "recaptcha_challenge_field": ""
            },
            headers={
                "x-requested-with": "XMLHttpRequest",
                "x-csrftoken": gen_csrftoken(),
                "Connection": "Keep-Alive",
                "Accept": "*/*",
                "Accept-Encoding": "gzip,deflate",
                "Accept-Language": "en-US",
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15",
            },
            proxies=self.public.proxies
        )
        try:
            return response.json()
        except JSONDecodeError as e:
            if "/login/" in response.url:
                raise ClientLoginRequired(e, response=response)
            raise ClientError(e, response=response)



import json
import time
import requests
import logging
from json.decoder import JSONDecodeError

from .utils import json_value
from .exceptions import (
    ClientError,
    ClientConnectionError,
    ClientNotFoundError,
    ClientJSONDecodeError,
    ClientForbiddenError,
    ClientBadRequestError,
    ClientGraphqlError,
    ClientThrottledError,
    ClientIncompleteReadError,
    ClientLoginRequired,
    GenericRequestError,
)


class PublicRequest:
    requests_count = 0
    PUBLIC_API_URL = "https://www.instagram.com/"
    GRAPHQL_PUBLIC_API_URL = "https://www.instagram.com/graphql/query/"
    request_logger = logging.getLogger("public_request")
    request_timeout = 1

    def __init__(self, *args, **kwargs):
        self.public = requests.Session()
        self.public.headers.update(
            {
                "Connection": "Keep-Alive",
                "Accept": "*/*",
                "Accept-Encoding": "gzip,deflate",
                "Accept-Language": "en-US",
                "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15",
            }
        )
        self.request_timeout = kwargs.pop("request_timeout", self.request_timeout)
        super().__init__(*args, **kwargs)

    def public_request(
        self,
        url,
        data=None,
        params=None,
        headers=None,
        return_json=False,
        retries_count=10,
        retries_timeout=10,
    ):
        kwargs = dict(
            data=data, params=params, headers=headers, return_json=return_json,
        )
        assert retries_count <= 10, "Retries count is too high"
        assert retries_timeout <= 600, "Retries timeout is too high"
        for iteration in range(retries_count):
            try:
                return self._send_public_request(url, **kwargs)
            except (
                ClientLoginRequired,
                ClientNotFoundError,
                ClientBadRequestError,
            ) as e:
                # Stop retries
                raise e
            except ClientError as e:
                msg = str(e)
                if (
                    isinstance(e, ClientConnectionError)
                    and "SOCKSHTTPSConnectionPool" in msg
                    and "Max retries exceeded with url" in msg
                    and "Failed to establish a new connection" in msg
                ):
                    raise e
                if retries_count > iteration + 1:
                    time.sleep(retries_timeout)
                else:
                    raise e
                continue

    def _send_public_request(
        self, url, data=None, params=None, headers=None, return_json=False
    ):
        self.requests_count += 1
        if headers:
            self.public.headers.update(headers)
        if self.request_timeout:
            time.sleep(self.request_timeout)
        try:
            if data is not None:  # POST
                response = self.public.data(url, data=data, params=params)
            else:  # GET
                response = self.public.get(url, params=params)

            expected_length = int(response.headers.get("Content-Length"))
            actual_length = response.raw.tell()
            if actual_length < expected_length:
                raise ClientIncompleteReadError(
                    "Incomplete read ({} bytes read, {} more expected)".format(
                        actual_length, expected_length
                    ),
                    response=response,
                )

            self.request_logger.debug(
                "public_request %s: %s", response.status_code, response.url
            )

            self.request_logger.info(
                "[%s] [%s] %s %s",
                self.public.proxies.get("https"),
                response.status_code,
                "POST" if data else "GET",
                response.url,
            )

            response.raise_for_status()
            return response.json() if return_json else response.text

        except JSONDecodeError as e:
            if "/login/" in response.url:
                raise ClientLoginRequired(e, response=response)

            self.request_logger.error(
                "Status %s: JSONDecodeError in public_request (url=%s) >>> %s",
                response.status_code,
                response.url,
                response.text,
            )
            raise ClientJSONDecodeError(
                "JSONDecodeError {0!s} while opening {1!s}".format(e, url),
                response=response,
            )
        except requests.HTTPError as e:
            if e.response.status_code == 403:
                raise ClientForbiddenError(e, response=e.response)

            if e.response.status_code == 400:
                raise ClientBadRequestError(e, response=e.response)

            if e.response.status_code == 429:
                raise ClientThrottledError(e, response=e.response)

            if e.response.status_code == 404:
                raise ClientNotFoundError(e, response=e.response)

            raise ClientError(e, response=e.response)

        except requests.ConnectionError as e:
            raise ClientConnectionError("{} {}".format(e.__class__.__name__, str(e)))

    def public_a1_request(self, endpoint, data=None, params=None, headers=None):
        url = self.PUBLIC_API_URL + endpoint.lstrip("/")
        if params:
            params.update({"__a": 1})
        else:
            params = {"__a": 1}

        response = self.public_request(
            url, data=data, params=params, headers=headers, return_json=True
        )
        try:
            return response["graphql"]
        except KeyError as e:
            error_type = response.get("error_type")
            if error_type == "generic_request_error":
                raise GenericRequestError(
                    json_value(response, "errors", "error", 0, default=error_type),
                    **response
                )
            raise e

    def public_graphql_request(
        self,
        variables,
        query_hash=None,
        query_id=None,
        data=None,
        params=None,
        headers=None,
    ):
        assert query_id or query_hash, "Must provide valid one of: query_id, query_hash"
        default_params = {"variables": json.dumps(variables, separators=(",", ":"))}
        if query_id:
            default_params["query_id"] = query_id

        if query_hash:
            default_params["query_hash"] = query_hash

        if params:
            params.update(default_params)
        else:
            params = default_params

        try:
            body_json = self.public_request(
                self.GRAPHQL_PUBLIC_API_URL,
                data=data,
                params=params,
                headers=headers,
                return_json=True,
            )

            if body_json.get("status", None) != "ok":
                raise ClientGraphqlError(
                    "Unexpected status '{}' in response. Message: '{}'".format(
                        body_json.get("status", None), body_json.get("message", None)
                    ),
                    response=body_json,
                )

            return body_json["data"]

        except ClientBadRequestError as e:
            message = None
            try:
                body_json = e.response.json()
                message = body_json.get("message", None)
            except JSONDecodeError:
                pass

            raise ClientGraphqlError(
                "Error: '{}'. Message: '{}'".format(e, message), response=e.response
            )


class TopSearchesPublic:
    def top_search(self, query):
        """Anonymous IG search request
        """
        url = "https://www.instagram.com/web/search/topsearch/"

        params = {
            "context": "blended",
            "query": query,
            "rank_token": 0.7763938004511706,
            "include_reel": "true",
        }

        response = self.public_request(url, params=params, return_json=True)
        return response


class HashtagPublic:
    def hashtag_info(self, hashtag, max_id=None):
        params = {"max_id": max_id} if max_id else None
        data = self.public_a1_request(
            "/explore/tags/{hashtag!s}/".format(**{"hashtag": hashtag}), params=params
        )
        return data["hashtag"]

    def hashtag_info_gql(self, hashtag, count=12, end_cursor=None):
        variables = {
            "tag_name": hashtag,
            "show_ranked": False,
            "first": int(count),
        }
        if end_cursor:
            variables["after"] = end_cursor

        data = self.public_graphql_request(
            variables, query_hash="f92f56d47dc7a55b606908374b43a314"
        )
        return data["hashtag"]

    def hashtag_top_feed(self, hashtag):
        data = self.hashtag_info(hashtag)
        return data["edge_hashtag_to_top_posts"]["edges"]

    def hashtag_related_hashtags(self, hashtag):
        data = self.hashtag_info(hashtag)
        return [
            item["node"]["name"]
            for item in data["edge_hashtag_to_related_tags"]["edges"]
        ]

    def hashtag_feed(self, hashtag, count=70, sleep=2):
        medias = []
        end_cursor = None

        while True:
            data = self.hashtag_info(hashtag, end_cursor)
            end_cursor = data["edge_hashtag_to_media"]["page_info"]["end_cursor"]
            edges = data["edge_hashtag_to_media"]["edges"]
            medias.extend(edges)

            if (
                not data["edge_hashtag_to_media"]["page_info"]["has_next_page"]
                or len(medias) >= count
            ):
                break

            time.sleep(sleep)

        return medias[:count]


class ProfilePublic:
    def location_feed(self, location_id, count=16, end_cursor=None):
        if count > 50:
            raise ValueError("Count cannot be greater than 50")

        variables = {
            "id": location_id,
            "first": int(count),
        }
        if end_cursor:
            variables["after"] = end_cursor

        data = self.public_graphql_request(
            variables, query_hash="1b84447a4d8b6d6d0426fefb34514485"
        )
        return data["location"]

    def profile_related_info(self, profile_id):
        variables = {
            "user_id": profile_id,
            "include_chaining": True,
            "include_reel": True,
            "include_suggested_users": True,
            "include_logged_out_extras": True,
            "include_highlight_reels": True,
            "include_related_profiles": True,
        }

        data = self.public_graphql_request(
            variables, query_hash="e74d51c10ecc0fe6250a295b9bb9db74"
        )
        return data["user"]



from .utils import json_value


def extract_media_v1(data):
    """Extract media from Private API
    """
    user = data["user"]
    location = data.get("location")
    if location:
        location = {"pk": int(location.get("pk")), "name": location.get("name")}
    video_url = ""
    if "video_versions" in data:
        video_url = sorted(
            data["video_versions"], key=lambda o: o["height"] * o["width"]
        ).pop()["url"]
    product_type = data.get("product_type", "")
    if data["media_type"] == 2 and not product_type:
        product_type = "feed"
    thumbnail_url = ''
    if 'image_versions2' in data:
        thumbnail_url = sorted(
            data["image_versions2"]["candidates"],
            key=lambda o: o["height"] * o["width"],
        ).pop()["url"]
    return {
        "pk": int(data["pk"]),
        "taken_at": int(data["taken_at"]),
        "id": data["id"],
        "media_type": data["media_type"],
        "product_type": product_type,
        "code": data["code"],
        "thumbnail_url": thumbnail_url,
        "location": location,
        "user": extract_user_short(user),
        "comment_count": int(data.get("comment_count") or 0),
        "like_count": int(data.get("like_count") or 0),  # the media just published has no like_count
        "caption_text": json_value(data, "caption", "text", default=""),
        "usertags": [
            extract_usertag(usertag)
            for usertag in data.get("usertags", {}).get("in", [])
        ],
        "video_url": video_url,
        "view_count": int(data.get('view_count') or 0),
        "video_duration": data.get('video_duration'),
        "title": data.get("title") or None,
        "resources": [
            extract_resource_v1(edge)
            for edge in data.get('carousel_media', [])
        ]
    }


def extract_media_gql(data):
    """Extract media from GraphQL
    """
    user = data["owner"]
    media_id = "%s_%s" % (data["id"], user["id"])
    if "full_name" in user:
        # for hashtag user contain {'id': '2041641294'}
        user = extract_user_short(user)
    else:
        user["pk"] = user.pop("id")
    location = data.get("location")
    if location:
        location = {"pk": int(location.get("id")), "name": location.get("name")}
    media_type = {"GraphImage": 1, "GraphVideo": 2, "GraphSidecar": 8}[data["__typename"]]
    product_type = data.get("product_type", "")
    video_url = ""
    if media_type == 2:
        video_url = data["video_url"]
        if not product_type:
            product_type = "feed"
    shortcode = ''
    if 'shortcode' in data:
        shortcode = data["shortcode"]
    return {
        "pk": int(data["id"]),
        "taken_at": int(data["taken_at_timestamp"]),
        "id": media_id,
        "media_type": media_type,
        "product_type": product_type,
        "code": shortcode,
        "thumbnail_url": sorted(
            data.get("display_resources", data.get('thumbnail_resources')),  # display_resources - user feed, thumbnail_resources - hashtag feed
            key=lambda o: o["config_width"] * o["config_height"],
        ).pop()["src"],
        "location": location,
        "user": user,
        "comment_count": json_value(data, "edge_media_to_comment", "count"),
        "like_count": json_value(data, "edge_media_preview_like", "count"),
        "caption_text": json_value(
            data, "edge_media_to_caption", "edges", 0, "node", "text", default=""
        ),
        "usertags": [
            extract_usertag(usertag['node'])
            for usertag in data.get("edge_media_to_tagged_user", {}).get("edges", [])
        ],
        "video_url": video_url,
        "view_count": int(data.get('video_view_count') or 0),
        "video_duration": data.get('video_duration'),
        "title": data.get("title") or None,
        "resources": [
            extract_resource_gql(edge['node'])
            for edge in data.get('edge_sidecar_to_children', {}).get('edges', [])
        ]
    }


def extract_resource_v1(data):
    video_url = ""
    if 'video_versions' in data:
        video_url = sorted(
            data["video_versions"], key=lambda o: o["height"] * o["width"]
        ).pop()["url"]
    thumbnail_url = sorted(
        data["image_versions2"]["candidates"],
        key=lambda o: o["height"] * o["width"],
    ).pop()["url"]
    return {
        "video_url": video_url,
        "thumbnail_url": thumbnail_url,
        "media_type": data['media_type'],
        "pk": int(data["pk"]),
        # "video_duration": data.get('video_duration'),
    }


def extract_resource_gql(data):
    media_type = {"GraphImage": 1, "GraphVideo": 2, "GraphSidecar": 8}[data["__typename"]]
    return {
        "video_url": data.get("video_url", ""),
        "thumbnail_url": data["display_url"],
        "media_type": media_type,
        "pk": int(data["id"]),
        # "view_count": int(data.get("video_view_count") or 0),
        # "shortcode": data["shortcode"],
        # "accessibility_caption": data.get("accessibility_caption")
    }


def extract_usertag(data):
    """Extract user tag
    """
    user = data['user']
    position = data.get('position')
    if not position:
        position = [data['x'], data['y']]
    return {
        "user": {
            "pk": int(user.get("id", user.get("pk"))),
            "username": user["username"],
            "full_name": user.get("full_name"),
            "profile_pic_url": user.get("profile_pic_url"),
            "is_verified": user.get("is_verified"),
        },
        "position": position
    }


def extract_user_short(data):
    """For Public GraphQL API
    """
    user_pk = data.get("id", data.get("pk"))
    assert user_pk, 'User without pk "%s"' % data
    return {
        "pk": int(user_pk),
        "username": data["username"],
        "full_name": data["full_name"],
        "is_private": data.get("is_private"),
        "profile_pic_url": data["profile_pic_url"],
        "is_verified": data.get("is_verified"),
        # "is_unpublished": data.get("is_unpublished"),
    }


def extract_user_gql(data):
    """For Public GraphQL API
    """
    return {
        "pk": int(data["id"]),
        "username": data["username"],
        "full_name": data["full_name"],
        "is_private": data["is_private"],
        "profile_pic_url": data["profile_pic_url"],
        "is_verified": data.get("is_verified"),
        "media_count": data["edge_owner_to_timeline_media"]["count"],
        "follower_count": data["edge_followed_by"]["count"],
        "following_count": data["edge_follow"]["count"],
        "biography": data["biography"],
        "external_url": data["external_url"],
        "is_business": data["is_business_account"],
    }


def extract_user_v1(data):
    """For Private API
    """
    return {
        "pk": int(data["pk"]),
        "username": data["username"],
        "full_name": data["full_name"],
        "is_private": data["is_private"],
        "profile_pic_url": data["profile_pic_url"],
        "is_verified": data.get("is_verified"),
        "media_count": data["media_count"],
        "follower_count": data["follower_count"],
        "following_count": data["following_count"],
        "biography": data["biography"],
        "external_url": data["external_url"],
        "is_business": data["is_business"],
    }



import json
from urllib.parse import urlparse

from .utils import InstagramIdCodec
from .exceptions import (
    ClientError,
    ClientNotFoundError,
    MediaNotFound,
)
from .decorators import check_login
from .extractors import extract_media_v1, extract_media_gql, extract_user_short


class Media:
    _medias_cache = {}  # pk -> object

    def media_id(self, media_id: str) -> str:
        """Return full media id
        Example: 2277033926878261772 -> 2277033926878261772_1903424587
        """
        media_id = str(media_id)
        if "_" not in media_id:
            assert media_id.isdigit(), (
                "media_id must been contain digits, now %s" % media_id
            )
            user = self.media_user(media_id)
            media_id = "%s_%s" % (media_id, user["pk"])
        return media_id

    @staticmethod
    def media_pk(media_pk: str) -> int:
        """Return short media id
        Example: 2277033926878261772_1903424587 -> 2277033926878261772
        """
        if "_" in str(media_pk):
            media_pk, _ = media_pk.split("_")
        return int(media_pk)

    def media_pk_from_code(self, code: str) -> int:
        """Return media_pk from code
        Example: B1LbfVPlwIA -> 2110901750722920960
        Example: B-fKL9qpeab -> 2278584739065882267
        """
        return InstagramIdCodec.decode(code)

    def media_pk_from_url(self, url: str) -> int:
        """Return media_pk from url
        Example: https://instagram.com/p/B1LbfVPlwIA/ -> 2110901750722920960
        Example: https://www.instagram.com/p/B-fKL9qpeab/?igshid=1xm76zkq7o1im -> 2278584739065882267
        """
        path = urlparse(url).path
        parts = [p for p in path.split("/") if p]
        return self.media_pk_from_code(parts.pop())

    def media_info_a1(self, media_pk: int, max_id=None) -> dict:
        media_pk = self.media_pk(media_pk)
        shortcode = InstagramIdCodec.encode(media_pk)
        """Use Client.media_info
        """
        params = {"max_id": max_id} if max_id else None
        data = self.public_a1_request(
            "/p/{shortcode!s}/".format(**{"shortcode": shortcode}), params=params
        )
        if not data.get("shortcode_media"):
            raise MediaNotFound(media_pk=media_pk, **data)
        return extract_media_gql(data["shortcode_media"])

    def media_info_gql(self, media_pk: int) -> dict:
        media_pk = self.media_pk(media_pk)
        shortcode = InstagramIdCodec.encode(media_pk)
        """Use Client.media_info
        """
        variables = {
            "shortcode": shortcode,
            "child_comment_count": 3,
            "fetch_comment_count": 40,
            "parent_comment_count": 24,
            "has_threaded_comments": False,
        }
        data = self.public_graphql_request(
            variables, query_hash="477b65a610463740ccdb83135b2014db"
        )
        if not data.get("shortcode_media"):
            raise MediaNotFound(media_pk=media_pk, **data)
        return extract_media_gql(data["shortcode_media"])

    def media_info_v1(self, media_pk: int) -> dict:
        try:
            result = self.private_request(f"media/{media_pk}/info/")
        except ClientNotFoundError as e:
            raise MediaNotFound(e, media_pk=media_pk, **self.last_json)
        except ClientError as e:
            if "Media not found" in str(e):
                raise MediaNotFound(e, media_pk=media_pk, **self.last_json)
            raise e
        return extract_media_v1(result["items"].pop())

    def media_info(self, media_pk: int, use_cache: bool = True) -> dict:
        """Return dict with media information
        """
        media_pk = self.media_pk(media_pk)
        if not use_cache or media_pk not in self._medias_cache:
            try:
                media = self.media_info_gql(media_pk)
            except Exception as e:
                if not isinstance(e, ClientError):
                    self.logger.exception(e)  # Register unknown error
                # Restricted Video: This video is not available in your country.
                # Or private account
                media = self.media_info_v1(media_pk)
            self._medias_cache[media_pk] = media
        return self._medias_cache[media_pk]

    @check_login
    def media_delete(self, media_id: str, media_type: str = '') -> bool:
        """Delete media
        Examples:
        https://i.instagram.com/api/v1/media/2277033926878261772_1903424587/delete/?media_type=PHOTO
        https://i.instagram.com/api/v1/media/2354534148830717883_1903424587/delete/?media_type=CAROUSEL
        """
        media_id = self.media_id(media_id)
        result = self.private_request(
            f"media/{media_id}/delete/", self.with_default_data({"media_id": media_id})
        )
        self._medias_cache.pop(self.media_pk(media_id), None)
        return result.get("did_delete")

    @check_login
    def media_edit(self, media_id: str, caption: str, title: str = "", usertags: list = []) -> bool:
        """Edit caption for media
        Example: https://i.instagram.com/api/v1/media/2154602296692269830_1903424587/edit_media/

        Video:
        {
            "caption_text": "Repost",
            "_csrftoken": "H8Rk6Ry2ffWcUSwWIBblVK4hHHII2RMk",
            "usertags": "{\"in\":[]}",
            "_uid": "8530598273",
            "device_id": "android-7d8ad96cc1b71922",
            "_uuid": "c642fece-8663-40d8-8ab7-112df0179e65",
            "is_carousel_bumped_post": "false",
            "container_module": "edit_media_info",
            "feed_position": "0",
            "location": "{}"
        }

        IGTV:
        {
            'igtv_ads_toggled_on': '0',
            'caption_text': 'TEXT',
            '_csrftoken': 'Bavik4rD52i0CvNqV1vDPNHBu4NcHQWB',
            '_uid': '1903424587',
            '_uuid': 'c642fece-8663-40d8-8ab7-112df0179e65',
            'title': 'zr+trip,+crimea,+feb+2017.+Edit+by+@milashensky'
        }
        """
        media_id = self.media_id(media_id)
        media = self.media_info(media_id)  # from cache
        usertags = [
            {"user_id": tag['user']['pk'], "position": tag['position']}
            for tag in usertags
        ]
        data = {
            "caption_text": caption,
            "container_module": "edit_media_info",
            "feed_position": "0",
            "location": "{}",
            "usertags": json.dumps({"in": usertags}),
            "is_carousel_bumped_post": "false",
        }
        if media["product_type"] == "igtv":
            if not title:
                try:
                    title, caption = caption.split("\n", 1)
                except ValueError:
                    title = caption[:75]
            data = {
                "caption_text": caption,
                "title": title,
                "igtv_ads_toggled_on": "0",
            }
        self._medias_cache.pop(self.media_pk(media_id), None)  # clean cache
        result = self.private_request(
            f"media/{media_id}/edit_media/", self.with_default_data(data),
        )
        return result

    def media_user(self, media_pk: int) -> dict:
        """Get user object
        """
        # return extract_user_short(
        #     self._media_info_a1(InstagramIdCodec.encode(media_pk))["owner"]
        # )
        return self.media_info(media_pk)["user"]

    def media_oembed(self, url: str) -> dict:
        """Return info about media and user by post URL
        Example: https://i.instagram.com/api/v1/oembed/?
            url=https://instagram.com/p/B1LbfVPlwIA?
            ig_mid=D68F44BF-2EDC-43AF-BB7F-D693BA0ABF05&
            utm_source=instagramweb
        Result: {
            "version": "1.0",
            "title": "",
            "author_name": "mind__flowers",
            "author_url": "https://www.instagram.com/mind__flowers",
            "author_id": 8572539084,
            "media_id": "2110901750722920960_8572539084",
            "provider_name": "Instagram",
            "provider_url": "https://www.instagram.com",
            "type": "rich",
            "width": 658,
            "height": null,
            "html": '...",
            "thumbnail_width": 640,
            "thumbnail_height": 799,
            "can_view": true,
        }
        """
        return self.private_request(f"oembed?url={url}")

    def media_comments(self, media_id: str) -> list:
        """Get list of comments for media
        Example: https://i.instagram.com/api/v1/media/2277659671519488169_8530598273/comments/?
        inventory_source=media_or_ad&
        analytics_module=comments_v2_feed_timeline&
        can_support_threading=true&
        is_carousel_bumped_post=false&
        feed_position=0
        """
        # TODO: to public or private
        media_id = self.media_id(media_id)
        max_id = None
        comments = []
        while True:
            try:
                result = self.private_request(
                    f"media/{media_id}/comments/", params={"max_id": max_id}
                )
                for comment in result["comments"]:
                    comment["user"] = extract_user_short(comment["user"])
                    comments.append(comment)
                if not result["has_more_comments"]:
                    break
                max_id = result["next_max_id"]
            except ClientNotFoundError as e:
                raise MediaNotFound(e, media_id=media_id, **self.last_json)
            except ClientError as e:
                if "Media not found" in str(e):
                    raise MediaNotFound(e, media_id=media_id, **self.last_json)
                raise e
        return comments

    @check_login
    def media_comment(self, media_id: str, text: str) -> int:
        """Comment media
        Example: {
            "user_breadcrumb": "u3QwgWYbhXMTf7nKhvpBYMyWxI9IScUmtUJ69TeTVwY=\nNCAxNDg5IDAgMTU4NjE5Mjg3MzI4NA==\n",
            "delivery_class": "organic",
            "idempotence_token": "9d68650c-bb63-4c09-a367-53d0bc177de4",
            "_csrftoken": "H8Rk6Ry2ffWcUSwWIBblVK4hHHII2RMk",
            "radio_type": "wifi-none",
            "_uid": "8530598273",
            "_uuid": "c642fece-8663-40d8-8ab7-112df0179e65",
            "comment_text": "Test",
            "is_carousel_bumped_post": "false",
            "container_module": "self_comments_v2_feed_contextual_self_profile",
            "feed_position": "0"
        }
        """
        media_id = self.media_id(media_id)
        result = self.private_request(
            f"media/{media_id}/comment/",
            self.with_action_data(
                {
                    "delivery_class": "organic",
                    "feed_position": "0",
                    "container_module": "self_comments_v2_feed_contextual_self_profile",  # "comments_v2",
                    "user_breadcrumb": self.gen_user_breadcrumb(len(text)),
                    "idempotence_token": self.generate_uuid(),
                    "comment_text": text,
                }
            ),
        )
        comment = result["comment"]
        comment["user"] = extract_user_short(comment["user"])
        return comment



import os
import shutil
import json
import time
import random
import requests
from uuid import uuid4
from PIL import Image
from urllib.parse import urlparse

from . import config
from .extractors import extract_media_v1
from .exceptions import PrivateError


class PhotoNotUpload(PrivateError):
    pass


class PhotoConfigureError(PhotoNotUpload):
    pass


class PhotoConfigureStoryError(PhotoConfigureError):
    pass


class DownloadPhoto:
    def photo_download(self, media_pk: int, folder: str = "/tmp") -> str:
        media = self.media_info(media_pk)
        assert media["media_type"] == 1, "Must been photo"
        filename = "{username}_{media_pk}".format(
            username=media["user"]["username"], media_pk=media_pk
        )
        return self.photo_download_by_url(media["thumbnail_url"], filename, folder)

    def photo_download_by_url(self, url: str, filename: str = "", folder: str = "/tmp") -> str:
        fname = urlparse(url).path.rsplit('/', 1)[1]
        filename = "%s.%s" % (filename, fname.rsplit('.', 1)[1]) if filename else fname
        filepath = os.path.join(folder, filename)
        response = requests.get(url, stream=True)
        if response.status_code == 200:
            with open(filepath, "wb") as f:
                response.raw.decode_content = True
                shutil.copyfileobj(response.raw, f)
        return os.path.abspath(filepath)


class UploadPhoto:
    def photo_rupload(
        self,
        filepath: str,
        upload_id: str = None,
        to_album: bool = False,
    ) -> tuple:
        """Upload photo to Instagram

        :param filepath:           Path to photo file (String)
        :param upload_id:          Unique upload_id (String). When None, then generate
                                   automatically. Example from video.video_configure

        :return: Tuple (upload_id, width, height)
        """
        assert isinstance(filepath, str), "Filepath must been string, now %s" % filepath
        upload_id = upload_id or str(int(time.time() * 1000))
        assert filepath, "Not specified filepath to photo"
        waterfall_id = str(uuid4())
        # upload_name example: '1576102477530_0_7823256191'
        upload_name = "{upload_id}_0_{rand}".format(
            upload_id=upload_id, rand=random.randint(1000000000, 9999999999)
        )
        rupload_params = {
            "retry_context": '{"num_step_auto_retry":0,"num_reupload":0,"num_step_manual_retry":0}',
            "media_type": "1",  # "2" if upload_id else "1",  # "2" when from video/igtv/album thumbnail, "1" - upload photo only
            "xsharing_user_ids": "[]",
            "upload_id": upload_id,
            "image_compression": json.dumps(
                {"lib_name": "moz", "lib_version": "3.1.m", "quality": "80"}
            ),
        }
        if to_album:
            rupload_params["is_sidecar"] = "1"
        photo_data = open(filepath, "rb").read()
        photo_len = str(len(photo_data))
        headers = {
            "Accept-Encoding": "gzip",
            "X-Instagram-Rupload-Params": json.dumps(rupload_params),
            "X_FB_PHOTO_WATERFALL_ID": waterfall_id,
            "X-Entity-Type": "image/jpeg",
            "Offset": "0",
            "X-Entity-Name": upload_name,
            "X-Entity-Length": photo_len,
            "Content-Type": "application/octet-stream",
            "Content-Length": photo_len,
        }
        response = self.private.post(
            "https://{domain}/rupload_igphoto/{name}".format(
                domain=config.API_DOMAIN, name=upload_name
            ),
            data=photo_data, headers=headers
        )
        self.request_log(response)
        if response.status_code != 200:
            self.logger.error(
                "Photo Upload failed with the following response: %s", response
            )
            last_json = self.last_json  # local variable for read in sentry
            raise PhotoNotUpload(response.text, response=response, **last_json)
        width, height = Image.open(filepath).size
        return upload_id, width, height

    def photo_upload(
        self,
        filepath: str,
        caption: str,
        upload_id: str = None,
        usertags: list = [],
        configure_timeout: int = 3,
        configure_handler=None,
        configure_exception=None
    ) -> dict:
        """Upload photo and configure to feed

        :param filepath:            Path to photo file (String)
        :param caption:             Media description (String)
        :param upload_id:           Unique upload_id (String). When None, then generate
                                        automatically. Example from video.video_configure
        :param usertags:            Mentioned users (List)
        :param configure_timeout:   Timeout between attempt to configure media (set caption, etc)
        :param configure_handler:   Configure handler method
        :param configure_exception: Configure exception class

        :return: Extracted media (Dict)
        """
        upload_id, width, height = self.photo_rupload(filepath, upload_id)
        for attempt in range(10):
            time.sleep(configure_timeout)
            if (configure_handler or self.photo_configure)(upload_id, width, height, caption, usertags):
                media = self.last_json.get("media")
                self.expose()
                return extract_media_v1(media)
        raise (configure_exception or PhotoConfigureError)(response=self.last_response, **self.last_json)

    def photo_configure(self, upload_id: str, width: int, height: int, caption: str, usertags: list) -> bool:
        """Post Configure Photo (send caption to Instagram)

        :param upload_id:  Unique upload_id (String)
        :param width:      Width in px (Integer)
        :param height:     Height in px (Integer)
        :param caption:    Media description (String)
        :param usertags:   Mentioned users (List)
        """
        usertags = [
            {"user_id": tag['user']['pk'], "position": tag['position']}
            for tag in usertags
        ]
        data = {
            "timezone_offset": "10800",
            "creation_logger_session_id": self.client_session_id,
            "multi_sharing": "1",
            "media_folder": "Camera",
            "source_type": "4",
            "caption": caption,
            "upload_id": upload_id,
            "device": self.device,
            "usertags": json.dumps({"in": usertags}),
            "edits": {
                "crop_original_size": [width * 1.0, height * 1.0],
                "crop_center": [0.0, 0.0],
                "crop_zoom": 1.0,
            },
            "extra": {"source_width": width, "source_height": height},
        }
        return self.private_request("media/configure/", self.with_default_data(data))

    def photo_upload_to_story(
        self,
        filepath: str,
        caption: str,
        upload_id: str = None,
        usertags: list = [],
        configure_timeout: int = 3,
    ) -> dict:
        """Upload photo and configure to story

        :param filepath:            Path to photo file (String)
        :param caption:             Media description (String)
        :param upload_id:           Unique upload_id (String). When None, then generate
                                        automatically. Example from video.video_configure
        :param usertags:            Mentioned users (List)
        :param configure_timeout:   Timeout between attempt to configure media (set caption, etc)

        :return: Extracted media (Dict)
        """
        return self.photo_upload(
            filepath, caption, upload_id, usertags,
            configure_timeout,
            configure_handler=self.photo_configure_to_story,
            configure_exception=PhotoConfigureStoryError
        )

    def photo_configure_to_story(self, upload_id: str, width: int, height: int, caption: str, usertags: list) -> bool:
        """Story Configure for Photo

        :param upload_id:  Unique upload_id (String)
        :param width:      Width in px (Integer)
        :param height:     Height in px (Integer)
        :param caption:    Media description (String)
        :param usertags:   Mentioned users (List)
        """
        timestamp = int(time.time())
        data = {
            "text_metadata": '[{"font_size":40.0,"scale":1.0,"width":611.0,"height":169.0,"x":0.51414347,"y":0.8487708,"rotation":0.0}]',
            "supported_capabilities_new": json.dumps(config.SUPPORTED_CAPABILITIES),
            "has_original_sound": "1",
            "camera_session_id": self.client_session_id,
            "scene_capture_type": "",
            "timezone_offset": "10800",
            "client_shared_at": str(timestamp - 5),  # 5 seconds ago
            "story_sticker_ids": "time_sticker_digital",
            "media_folder": "Camera",
            "configure_mode": "1",
            "source_type": "4",
            "creation_surface": "camera",
            "imported_taken_at": (timestamp - 3 * 24 * 3600),  # 3 days ago
            "caption": caption,
            "capture_type": "normal",
            "rich_text_format_types": '["default"]',
            "upload_id": upload_id,
            "client_timestamp": str(timestamp),
            "device": self.device,
            "implicit_location": {
                "media_location": {
                    "lat": 44.64972222222222,
                    "lng": 33.541666666666664
                }
            },
            "edits": {
                "crop_original_size": [width * 1.0, height * 1.0],
                "crop_center": [0.0, 0.0],
                "crop_zoom": 1.0,
            },
            "extra": {"source_width": width, "source_height": height},
        }
        if usertags:
            mentions = [
                {
                    "x": 0.5002546, "y": 0.8583542, "z": 0,
                    "width": 0.4712963, "height": 0.0703125, "rotation": 0.0,
                    "type": "mention", "user_id": str(tag['user']['pk']), "is_sticker": False, "display_type": "mention_username"
                } for tag in usertags
            ]
            data["tap_models"] = data["reel_mentions"] = json.dumps(mentions)
        return self.private_request("media/configure_to_story/", self.with_default_data(data))



import os.path
from setuptools import setup, find_packages

base = os.path.abspath(os.path.dirname(__file__))

setup(
    name='instagrapi',
    version='1.0.4',
    author='Mikhail Andreev',
    author_email='x11org@gmail.com',
    license='MIT',
    url='https://github.com/adw0rd/instagrapi',
    install_requires=[
        'pytz==2020.1',
        'requests==2.24.0',
        'moviepy==1.0.3',
        'Pillow==7.2.0'
    ],
    # test_requires=[],
    keywords='instagram private api',
    description='Fast and effective Instagram Private API wrapper (public+private requests and challenge resolver)',
    long_description=open(os.path.join(base, 'README.md')).read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires=">=3",
    package_data={'': ["requirements.txt"]},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
    ],
)



import os
import tempfile

from moviepy.editor import TextClip, CompositeVideoClip, VideoFileClip, ImageClip
from PIL import Image, ImageFont, ImageDraw


class StoryBuilder:
    width = 720
    height = 1280

    def __init__(self, filepath: str, caption: str = "", usertags: list = []):
        self.filepath = filepath
        self.caption = caption
        self.usertags = usertags
        self.bgpath = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "assets", "story", "bg1.png"
        )

    def build_clip(self, clip, max_duration):
        background = ImageClip(self.bgpath)
        clip_left = (self.width - clip.size[0]) / 2
        clip_top = (self.height - clip.size[1]) / 2
        if clip_top > 90:
            clip_top -= 50
        clip = clip.set_position((clip_left, clip_top))
        caption = self.caption
        tag = None
        if self.usertags:
            tag = self.usertags[0]
            caption = "@%s" % tag["user"]["name"]
        text_clip = TextClip(caption, color="white", font="Arial", kerning=-1, fontsize=100, method="label")
        text_clip_left = (self.width - 600) / 2
        text_clip_top = clip_top + clip.size[1] + 50
        offset = (text_clip_top + text_clip.size[1]) - self.height
        if offset > 0:
            text_clip_top -= offset + 90
        text_clip = text_clip.resize(width=600).set_position((text_clip_left, text_clip_top)).fadein(3)
        usertags = []
        if tag:
            tag['x'] = 0.49892962  # approximately center
            tag['y'] = (text_clip_top + text_clip.size[1] / 2) / self.height
            tag['width'] = text_clip.size[0] / self.width
            tag['height'] = text_clip.size[1] / self.height
            usertags = [tag]
        duration = max_duration
        if max_duration and clip.duration and max_duration > clip.duration:
            duration = clip.duration
        destination = tempfile.mktemp('.mp4')
        CompositeVideoClip(
            [background, clip, text_clip], size=(self.width, self.height)
        ).set_fps(24).set_duration(duration).write_videofile(destination, codec='libx264', audio=True, audio_codec='aac')
        return {
            'usertags': usertags,
            'filepath': destination
        }

    def video(self, max_duration: int = 0):
        clip = VideoFileClip(self.filepath, has_mask=True)
        return self.build_clip(clip, max_duration)

    def photo(self, max_duration: int = 0):
        clip = ImageClip(self.filepath).resize(width=self.width)
        return self.build_clip(clip, max_duration or 14)

    def photo_pil(self):
        """Unfinished
        """
        background = Image.open(self.bgpath)
        txt = Image.new('RGBA', background.size, (255, 255, 255, 0))
        fnt = ImageFont.load_default()
        caption = self.caption
        tag = None
        if self.usertags:
            tag = self.usertags[0]
            caption = "@%s" % tag["user"]["name"]
        draw = ImageDraw.Draw(txt)
        draw.text((360, 640), caption, font=fnt, fill=(255, 255, 255, 128))
        out = Image.alpha_composite(background, txt)
        destination = tempfile.mktemp('.jpg')
        # out.show()
        out.convert('RGB').save(open(destination, 'w'))
        return {
            'usertags': self.usertags,
            'filepath': destination
        }
